package org.example

import scala.collection.parallel.immutable.ParRange

class RangeScala(val start: Int, val end: Int, val step: Int) extends IndexedSeq[Int]
	   with collection.CustomParallelizable[Int, ParRange]
	   with Serializable{
  
  def isInclusive = false
  
  def by(step: Int): RangeScala = copy(start, end, step)
  protected def copy(start: Int, end: Int, step: Int): RangeScala = new RangeScala(start, end, step)
  
  @inline
  final def apply(idx: Int): Int = {
	if (idx < 0 || idx >= length) throw new IndexOutOfBoundsException(idx.toString)
	locationAfterN(idx)
  }
  
  private def locationAfterN(n: Int) = start + (step * n)
  override def length: Int = numRangeElements
  private lazy val numRangeElements: Int = Range.count(start, end, step, isInclusive)
  
    final override def reverse: RangeScala =
	    if (length > 0) new RangeScala.Inclusive(last, start, -step)
	    else this

  // has to be public, due to access in foreach (if private then foreach can't be inlined)
  val numSteps = preproc(start, end, step, isInclusive)

  private[this] def preproc(start: Long, end: Long, step: Long, isInclusive: Boolean): Int = {
    val upward  = (start < end)
    if (step == 0) -1
    else if (start == end) if (isInclusive) 1 else 0
    else if (upward != (step  > 0)) 0
    else {
      val diff      = end - start
      val jumps     = diff / step
      val remainder = diff % step
      val longCount = jumps + (
        if (!isInclusive && 0 == remainder) 0 else 1
      )

      /** The edge cases keep coming.  Since e.g.
       *    Long.MaxValue + 1 == Long.MinValue
       *  we do some more improbable seeming checks lest
       *  overflow turn up as an empty range.
       */
      // The second condition contradicts an empty result.
      val isOverflow = (longCount == 0) && (((start + step) < end) == upward)

      if (longCount > scala.Int.MaxValue || longCount < 0L || isOverflow) -1
      else longCount.toInt
    }
  }

  /** @note Making foreach run as fast as a while loop is a challenge.
   */
  @inline final override def foreach[U](f: Int => U) {
    var numIters = numSteps
    if(numIters == -1) {
      if (step == 0) throw new IllegalArgumentException("step cannot be 0.")
      else {
        val word  = if (isInclusive) "to" else "until"
        val descr = List(start, word, end, "by", step) mkString " "

        throw  new IllegalArgumentException(descr + ": seqs cannot contain more than Int.MaxValue elements.")
      }
    }
    var i = start
    while ( numIters > 0 ) {
      (f.asInstanceOf[Int => Unit]).apply(i)
      i += step
      numIters -= 1
    }
  }
  
}

object RangeScala {
    private[example] val MAX_PRINT = 512  // some arbitrary value
		
	class Inclusive(start: Int, end: Int, step: Int) extends RangeScala(start, end, step) {
	//    override def par = new ParRange(this)
	    override def isInclusive = true
	    override protected def copy(start: Int, end: Int, step: Int): RangeScala = new Inclusive(start, end, step)
	}
	
	  /** Make a range from `start` until `end` (exclusive) with given step value.
	   * @note step != 0
	   */
	  def apply(start: Int, end: Int, step: Int): RangeScala = new RangeScala(start, end, step)
	
	  /** Make an range from `start` to `end` inclusive with step value 1.
	   */
	  def apply(start: Int, end: Int): RangeScala = new RangeScala(start, end, 1) 
	
	  /** Make an inclusive range from start to end with given step value.
	   * @note step != 0
	   */
	  @inline def inclusive(start: Int, end: Int, step: Int): RangeScala.Inclusive = new Inclusive(start, end, step)
	
	  /** Make an inclusive range from start to end with step value 1.
	   */
	  @inline def inclusive(start: Int, end: Int): RangeScala.Inclusive = new Inclusive(start, end, 1) 
			 	 
	  // As there is no appealing default step size for not-really-integral ranges,
	  // we offer a partially constructed object.
	  class Partial[T, U](f: T => U) {
	    def by(x: T): U = f(x)
	  }
	 
}